use serde::{Deserialize, Serialize};

use crate::{fs_util::{dir_exists, file_exists, join, mkdir, read_dir}, Id, Inventory, Kind, Valid, logger::{trace, debug}};

#[derive(Debug, PartialEq, Clone, Serialize, Deserialize, Default)]
pub struct Object {
    // Required Fields
    pub id: Id,
    pub kind: Kind,
    pub name: String,

    // Optional Fields
    pub password: String,
    pub owner: Id,
    pub location: Id,
    pub health: u64,
    pub max_health: u64,
    pub magic: u64,
    pub max_magic: u64,
    pub build: u64,
    pub max_build: u64,
    pub inv: Inventory
}

impl Valid for Object {
    fn valid(&self) -> bool {
        if self.kind == Kind::Player {
            if self.password.is_empty() {
                debug!("Object.valid() > Player Missing Object.password");
                return false;
            }
        }
        self.id.valid() && self.kind.valid() && !self.name.is_empty()
    }
}

impl Object {
    pub fn set_password(&mut self, password: &String) -> bool {
        let hash = bcrypt::hash(password, bcrypt::DEFAULT_COST);
        match hash {
            Ok(hash) => {
                self.password = hash;
                true
            }
            Err(e) => {
                debug!("Object.set_password('{}') > {}", password, e);
                false
            }
        }
    }
    pub fn verify_password(&self, password: &String) -> bool {
        let verify = bcrypt::verify(password, self.password.as_str());
        match verify {
            Ok(ok) => ok,
            Err(e) => {
                debug!("Object.verify_password() > {}", e);
                false
            }
        }
    }

    pub fn load_file(filename: &String) -> Option<Object> {
        let p = crate::fs_util::join(&format!("{}", filename));
        if !file_exists(&p) {
            trace!("Object.load_file('{}') > '{}' Doesn't exist", filename, p);
            return None;
        }
        let cont = crate::fs_util::read_file(&p);
        if cont.is_empty() {
            debug!("Object.load_file('{}') > Failed reading '{}'", filename, p);
            return None;
        }
        Object::load(&cont)
    }
    pub fn load(cont: &String) -> Option<Object> {
        let obj: Result<Object, serde_json::Error> = serde_json::from_str(cont.as_str());
        match obj {
            Ok(obj) => Some(obj),
            Err(e) => {
                debug!("Object.load() > {}", e);
                None
            }
        }
    }
    pub fn load_by_id(id: Id) -> Option<Object> {
        Object::load_file(&format!("data/objects/{}.json", id))
    }
    pub fn load_player_by_name(name: &String) -> Option<Object> {
        let p = join(&"data/objects".to_string());
        let dir = read_dir(&p, false);
        if dir.is_empty() {
            if !dir_exists(&p) {
                trace!("Object.load_player_by_name('{}') > data/objects directory doesn't exist, creating", name);
                mkdir(&p);
            }
            return None;
        }
        for entry in dir.iter() {
            //trace!("Object.load_player_by_name('{}') > Trying '{}'", name, entry);
            let obj = Object::load_file(entry);
            match obj {
                Some(obj) => {
                    if obj.name == *name && obj.kind == Kind::Player {
                        trace!("Object.load_player_by_name('{}') > Found in '{}'", name, entry);
                        return Some(obj);
                    }
                }
                None => {
                    trace!("Object.load_player_by_name('{}') > Errors in '{}'", name, entry);
                }
            }
        }
        trace!("Object.load_player_by_name('{}') > Not found", name);
        None
    }
    pub fn save(&mut self) -> bool {
        if !self.valid() {
            debug!("Object.save() > Not valid, won't save");
            return false;
        }
        let p = crate::fs_util::join(&format!("data/objects/{}.json", self.id));
        let dir = crate::fs_util::join(&"data/objects".to_string());
        if !crate::fs_util::dir_exists(&dir) {
            debug!("Object.save() > Missing '{}' directory, creating...", dir);
            if !crate::fs_util::mkdir(&dir) {
                debug!("Object.save() > Failed creating '{}' directory", dir);
            }
        }
        let cont = serde_json::to_string(self);
        match cont {
            Ok(cont) => {
                if !crate::fs_util::write_file(&p, &cont) {
                    debug!("Object.save() > Failed writing '{}'", p);
                    return false;
                }
                true
            }
            Err(e) => {
                debug!("Object.save() > Failed serializing, {}", e);
                false
            }
        }
    }
}
