use std::collections::HashMap;

use crate::logger::trace;
use serde::{Deserialize, Serialize};

use crate::{fs_util::{dir_exists, join, mkdir, read_dir}, object::Object};

pub trait Valid {
    fn valid(&self) -> bool;
}

pub type Id = u64;

impl Valid for Id {
    fn valid(&self) -> bool {
        *self != 0
    }
}

pub fn next_id() -> Id {
    let p = join(&"data/objects".to_string());
    let dir = read_dir(&p, false);
    if dir.is_empty() {
        if !dir_exists(&p) {
            mkdir(&p);
        }
        trace!("next_id() > Empty dir > 1");
        return 1;
    }
    let mut last: Id = 1;
    for entry in dir.iter() {
        //eprintln!("Trying '{}'", entry);
        let obj = Object::load_file(entry);
        match obj {
            Some(obj) => {
                if last < obj.id {
                    last = obj.id;
                }
            }
            None => {
                trace!("next_id() > Errors in '{}'", entry);
            }
        }
    }
    trace!("next_id() > {}", last + 1);
    last + 1
}

#[derive(Debug, PartialEq, Clone, Serialize, Deserialize)]
pub enum Kind {
    None,

    Player,
    Room,
    Item,
    Creature,
    Vehicle
}

impl Valid for Kind {
    fn valid(&self) -> bool {
        !matches!(self, Kind::None)
    }
}

impl Default for Kind {
    fn default() -> Self {
        Kind::None
    }
}

#[derive(Debug, PartialEq, Clone, Serialize, Deserialize, Default)]
pub struct Inventory {
    data: HashMap<Id, u64>,
    pub max_slots: u64
}

impl Valid for Inventory {
    fn valid(&self) -> bool {
        self.max_slots != 0
    }
}

impl Inventory {
    pub fn slots_used(&self) -> u64 {
        self.data.len() as u64
    }
    pub fn slots_left(&self) -> u64 {
        let used = self.slots_used();
        if used > self.max_slots {
            return 0;
        }
        self.max_slots - used
    }
    pub fn has(&self, id: Id) -> bool {
        let amt = self.data.get(&id);
        match amt {
            Some(amt) => *amt != 0,
            None => false
        }
    }
    pub fn amount(&self, id: Id) -> u64 {
        let amt = self.data.get(&id);
        match amt {
            Some(amt) => *amt,
            None => 0
        }
    }
    pub fn has_amount(&self, id: Id, amt: u64) -> bool {
        let amount = self.amount(id);
        amount >= amt
    }
    pub fn add(&mut self, id: Id, amt: u64) -> bool {
        let amount = self.amount(id);
        if amount != 0 {
            self.data.insert(id, amount + amt);
            return true;
        }
        if self.slots_left() == 0 {
            return false;
        }
        self.data.insert(id, amt);
        true
    }
    pub fn add_one(&mut self, id: Id) -> bool {
        self.add(id, 1)
    }
    pub fn take(&mut self, id: Id, amt: u64) -> bool {
        let amount = self.amount(id);
        if amount == 0 {
            return false;
        }
        if amount > amt {
            self.data.insert(id, amount - amt);
            true
        } else if amount == amt {
            self.data.remove(&id);
            true
        } else {
            false
        }
    }
    pub fn take_one(&mut self, id: Id) -> bool {
        self.take(id, 1)
    }
    pub fn take_all(&mut self, id: Id) -> u64 {
        let amt = self.data.remove(&id);
        match amt {
            Some(amt) => amt,
            None => 0
        }
    }
}
